----------------------------------------------------------------------------------
--
--	Fast 65C02 with local RAM
--
-- Replaces a Rockwell, Ricoh or compatible 65C02 in a host system.
-- Needs to know where I/O is mapped in the host system memory space
-- (via <map> inputs and pre-defined memory maps in the <decoder> module).
-- Copies all non-I/O memory into in-board RAM at startup,
-- executes programs fast from internal RAM, accesses external bus with original phi0 speed.
--
--	(C) 2020 Juergen Mueller, juergen@e-basteln.de
-- Based on a concept by Roland Langfeld
--
----------------------------------------------------------------------------------
--
-- Based on original 6502 "Arlet 6502 Core" by Arlet Ottens 
-- Support for 65C02 instructions and addressing modes by David Banks and Ed Spittles
-- (C) Arlet Ottens, <arlet@c-scape.nl>
-- (C) 2016 David Banks and Ed Spittles
--
-- Feel free to use this code in any project (commercial or not), as long as you
-- keep this message, and the copyright notice. This code is provided "as is", 
-- without any warranties of any kind.
--
----------------------------------------------------------------------------------
--
--	jm 14.4.2020	CPU & RAM only
-- jm 18.4.2020	add external bus interface. RDY input not supported yet, untested.
-- jm 18.4.2020	add initialization state machine, untested.
-- jm 19.4.2020	RDY support, move clock into separate module (with DCM for tests on Numato -2 FPGA @ 80MHz)
-- jm 19.4.2020	add dummy ROM module for testing
-- jm 20.4.2020	data out during phi2 only 
-- jm 24.4.2020	convert to Arlet Ottens' 65C02 core. SYNC output added, /SO input missing.
-- jm 28.4.2020	test synthesis for final PCB (test output mapping). 80 MHz clock defined, max. is 82 MHz (-3 FPGA).
-- jm  5.5.2020	removed timing optimizations for Apple II (asymmetrical clocks, early databus read via ext_d_prev)
-- jm  7.5.2020	timing tweaks for Mephisto: phi2 really symmetrical again now, adr/rw hold = 2*clk, data hold = 6*clk.
--	jm	 7.5.2020	test: hold data even longer, until phi2 goes high again
-- jm  8.5.2020	host-specific "idle" address from decoder module; shortened the data hold time again.
-- jm  8.5.2020	adr hold time = 3*clk, data hold time = 7*clk
-- jm 23.5.2020	RDY handling in external bus FSM added
-- jm 24.5.2020	adr hold time = 3*clk, data hold time = 4*clk -- is this better for Mephisto?
-- jm 31.5.2020	for Rev A custom PCB: positive ext_drive_o replaces inverted ext_en_n_o
-- jm 31.5.2020	re-implement Apple II specific changes:
--						- asymmetrical phi1, phi2 clock;
--						- ext_d_prev buffer for data input !!!
--						- back to 75 MHz; 
--						- back to original (incorrect) ram_en.
-- jm 10.6.2020	dial back the Apple II specifics. 
--						- asymm. clock not needed.
--						- ahold=3, dhold=4 or 7 works (@75MHz)
--						- fast ram_en works
--						- ext_d_prev IS required!
--	jm 10.6.2020	try external memory access through BRAM - faster, due to lack of data mux? no, only 95 MHz. 
--						Also, logic does not work (65c02 core moves on to next address upon external read cycles)
-- jm 11.6.2020	try RAM organized as 64 kbit * 1 - faster? yes, 99.8 MHz, but not reproducible.
--						Also, would prevent use of bit 9 to mark bytes to be written back.
-- jm 13.6.2020	both changes rolled back, switch to 100 MHz PLL, placement restriction for 65c02_mmux >> 100 MHz
-- jm 14.6.2020	integrate TinyFPGA USB/SPI bridge
-- jm 15.6.2020	fix SYNC output, add /SO input -- not tested yet
-- jm 16.6.2020	ext_drive_n_o is inverted for Rev C board
-- jm 21.6.2020	read buffer enable (ext_dread_n_o) for Rev D board, workaround via USB- for Rev C
--	jm 27.6.2020	Adjust timing according to measurements on real Ricoh R65C02G (4 MHz):
--						Addr Hold & R/W Hold 40ns, Data Hold 50 ns.
--	jm 10.7.2020	remove USB functionality (move to separate boot loader configuration)
-- jm 10.7.2020	new decoder for external addresses, tries to detect time-critical code 
--						based on I/O addresses, ends time-critical segment on RTS/RTI. Untested!
-- jm 26.7.2020	remove ext_d_prev; not required by Apple II anymore
-- jm 14.8.2020	JTAG TDI=0 overrides DIP switch setting
--
-- to do: Debug Automatic slowdown
--	/ Ensure that read access in slow mode gets the right data byte! (data still on internal bus? cpu_adr still valid at RAM?)
--	/ Ensure that no incorrect external write occurs in slow mode!
-- / Timeout to switch back to fast mode after top-level I/O (e.g. bootstraps).
--	/ Debug Apple DOS boot
-- * Base RTS/JSR control on stack level instead, to deal with PHA/PHA/RTS indirect jumps?
--
----------------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

library work;

library unisim;
use unisim.vcomponents.all;


entity Spartan_65C02 is
	port (
	
		----------------------------------------------------
		-- replica connections
		----------------------------------------------------

		clk_i   			: in  	std_logic;								-- FPGA clock
		map_n_i			: in		std_logic_vector (1 to 4);			-- select memory model
		led				: out		std_logic;								-- status LED
		test				: out		std_logic_vector (1 downto 0);	-- test outputs (DIP pins 35, 36)
		
		----------------------------------------------------
		-- external (slow) 65c02 connections
		----------------------------------------------------
		
		-- clock
		ext_phi0_i  	: in     std_logic;
		ext_phi1_o  	: out    std_logic;
		ext_phi2_o  	: out    std_logic;
		
		-- address and data bus
		ext_d_io      	: inout	std_logic_vector (7 downto 0);
		ext_a_o       	: out    std_logic_vector (15 downto 0);
		ext_wr_n_o    	: out    std_logic;								-- Read /Write, to host system
		ext_dwrite_n_o	: out		std_logic;								-- Rev C&D: /Write during Phi2=H only, low enables output driver
		ext_dread_n_o	: out		std_logic;								-- Rev C&D: /Read, low enables input buffer

		-- reset and interrupts
		ext_rst_n_i 	: in     std_logic;
		ext_irq_n_i   	: in     std_logic;
		ext_nmi_n_i   	: in     std_logic;

		-- READY and SYNC, Set Overflow
		ext_sync_o    	: out    std_logic;
		ext_rdy_i     	: in     std_logic;
		ext_so_n_i    	: in     std_logic
		
   );
end Spartan_65C02;

architecture BEHAVIORAL of Spartan_65C02 is


-------------------------------------------------
-- cpu signals
-------------------------------------------------

	signal clk_buf		: std_logic;							-- buffered external clock
	signal map_i		: std_logic_vector (1 to 4);		-- inverted memory map selectors (active high)
	signal map_enable	: std_logic;							-- can override DIP settings via TDI=0
	
	 -- connect internal 65C02 Core
	 --
	signal clk   		: std_logic;
	signal cpu_irq    : std_logic := '0';
	signal cpu_nmi    : std_logic := '0';
	signal cpu_rdy    : std_logic := '1';
	signal cpu_rst 	: std_logic := '0';
	signal cpu_so     : std_logic := '0';
	signal cpu_adr    : std_logic_vector (15 downto 0);
	signal cpu_dout   : std_logic_vector (7 downto 0);
	signal cpu_din		: std_logic_vector (7 downto 0);	-- multiplexed data
	signal cpu_sync	: std_logic	:= '0';					-- high while instruction data are read
	signal cpu_wr		: std_logic;
	
	-- Track slow mode.
	--
	-- Slow mode is entered when the decoder unit detects access to a "time critical" I/O address.
	-- Slow mode ends when the processor returns upwards from the level where such access was first detected,
	-- via an RTS or RTI. JSR/RTS nesting depth is counted to keep track.
	-- Slow mode also ends after a timeout (external bus cycles counted via slow_timeout),
	-- to deal with cases where time-critical IO is accessed by top level code, e.g. during startup.

	signal is_rts		: std_logic;							-- executing RTS or RTI
	signal is_jsr		: std_logic;							-- executing BRK, IRQ, NMI
	signal slow_mode	: std_logic := '0';					-- currently executing time-critical code section
	signal rts_level	: natural range 0 to 255 := 0;	-- subroutine/irq nesting level
	signal slow_timeout: natural range 0 to 16#FFFFF#;	-- timeout counter, counts external bus cycles in slow mode
	
	-- synchronize asynchronous inputs
	--
	signal meta_phi0, 	sync_phi0	: std_logic ;
	signal meta_irq_n,	sync_irq_n	: std_logic ;
	signal meta_nmi_n,	sync_nmi_n  : std_logic ;
	signal meta_rdy, 		sync_rdy    : std_logic ;
	signal meta_rst_n,	sync_rst_n 	: std_logic ;
	signal meta_so_n, 	sync_so_n 	: std_logic ;
	
	attribute ASYNC_REG : string;
	attribute ASYNC_REG of meta_phi0		: signal is "TRUE";
	attribute ASYNC_REG of sync_phi0		: signal is "TRUE";
	attribute ASYNC_REG of meta_irq_n	: signal is "TRUE";
	attribute ASYNC_REG of sync_irq_n	: signal is "TRUE";
	attribute ASYNC_REG of meta_nmi_n	: signal is "TRUE";
	attribute ASYNC_REG of sync_nmi_n	: signal is "TRUE";
	attribute ASYNC_REG of meta_rdy		: signal is "TRUE";
	attribute ASYNC_REG of sync_rdy		: signal is "TRUE";
	attribute ASYNC_REG of meta_rst_n	: signal is "TRUE";
	attribute ASYNC_REG of sync_rst_n	: signal is "TRUE";
	attribute ASYNC_REG of meta_so_n		: signal is "TRUE";
	attribute ASYNC_REG of sync_so_n		: signal is "TRUE";	
	
	-- edge detection and delayed phi2 edge
	--
	signal prev0_phi2		: std_logic;				-- delayed status of phi2, to determine hold times
	signal prev1_phi2		: std_logic;				-- ### convert to a proper shift register with selectable taps?
	signal prev2_phi2		: std_logic;
	signal prev3_phi2		: std_logic;
	signal prev4_phi2		: std_logic;
	signal prev5_phi2		: std_logic;
	signal prev6_phi2		: std_logic;
	signal prev7_phi2		: std_logic;
	
	signal edgep_phi2		: std_logic := '0';		-- positive edge on phi2
	signal edgen_phi2		: std_logic := '0';		-- negative edge on phi2
	signal edgen_phi2_ahold	: std_logic := '0';	-- addresses and data change, some cycles after negative edge on phi2
	signal edgen_phi2_dhold	: std_logic := '0';	-- data hold time ends, some more cycles after negative edge on phi2
	
	-- map between internal CPU and external bus
	--
	signal is_ext		 	: std_logic;								-- current address is external (peripheral)
	signal was_ext			: std_logic;								-- prior address was external
	signal is_slow			: std_logic;								-- operating in slow mode (timing-critical code assumed)
	signal ram_en			: std_logic;								-- enable internal RAM
	signal idle_page		: std_logic_vector (7 downto 0);		-- page address to put on bus while cpu works internally

	signal ext_drive_buf	: std_logic;								-- shadow copy of ext_drive_o, enables internal bidir data bus driver 
	signal ext_a_buf		: std_logic_vector (15 downto 0);	-- shadow copy of ext_a_o

	signal mem_dout		 	: std_logic_vector (7 downto 0);	-- data from internal memory
	signal ext_d_i		 	: std_logic_vector (7 downto 0);		-- data from external bus
	signal ext_d_buf	 	: std_logic_vector (7 downto 0);		-- previous clk's data from external bus, latched at Phi2 > low
	signal ext_d_o		 	: std_logic_vector (7 downto 0);		-- data to external bus, latched
	signal rdy_int		 	: std_logic := '1';						-- RDY generated by external bus interface to stop CPU
	signal rst_n_int	 	: std_logic := '0';						-- Power-on reset of internal CPU, keep low during init loop
	
	signal cpu_sync_latched	: std_logic;							-- sync signal latched for external bus access
	signal cpu_wr_latched	: std_logic;							-- write signal latched for external bus access
	signal cpu_adr_latched	: std_logic_vector (15 downto 0);-- address latched for external bus access
	signal cpu_dout_latched	: std_logic_vector (7 downto 0);	-- data latched for external bus access
	
	signal led_latch		: std_logic;								-- decouple LED output (timing)
	signal test_latch		: std_logic_vector (1 downto 0);		-- decouple test outputs (timing)
	
	-- state machine for initialization and external (slow) bus accesses
	--
	type bus_states is						-- state of external databus
		(
			bus_init_waitreset,				-- wait for power-on reset to be released
			bus_init_waitwrite,				-- wait until initial write cycle (host initialization) complete
			bus_init_wait,						-- looping through addresses - wait for delayed low edge of phi2
			bus_init_active,					-- looping through addresses - bus currently active
			bus_idle,							-- no access, cpu is running from internal RAM
			bus_wait_ahold,					--	wait for end of adr hold time (delayed low edge of phi2)
			bus_wait_dhold,					--	wait for end of data hold time (delayed low edge of phi2)
			bus_wait_write,					-- wait for trigger to write to data bus (phi2 rising edge)
			bus_wait_read						-- wait for trigger to read from data bus (phi2 falling edge)
		);
	signal bus_state		: bus_states := bus_init_waitreset;
	signal new_cycle		: std_logic;	-- sub-state of bus_wait_ahold: first clk tick
	
	-- initialization loop:
	-- read all RAM & ROM from host system, write to internal RAM
	signal a_init			: std_logic_vector (15 downto 0);	-- cycle through addresses via dedicated counter
	signal is_ext_init	: std_logic;	-- is this a peripheral address? then don't read
	signal wr_init 		: std_logic;	-- write trigger for internal RAM
	signal en_init			: std_logic;	-- RAM port B is enabled during initialization only
	signal a_init_write	: std_logic_vector (15 downto 0);	-- address for host initialization prior to reading RAM/ROM
	signal d_init_write	: std_logic_vector (7 downto 0);		-- data for host initialization 
			
begin


---------------------------------------------------------------------
-- instantiated components
---------------------------------------------------------------------

	BUF_clk_buf  : BUFIO2 port map ( I => clk_i, DIVCLK => clk_buf);

	-- clock for custom DIP40 PCB, Grade -3 FPGA
	clock_main_inst: entity work.clock_main
	port map ( 
		clk_in1 	=> clk_buf,				-- external clock input, 25 MHz
		clk_out1	=> clk					-- buffered system clock, 100 MHz
	);


	inst_65c02: entity work.cpu_65c02 
	port map(
		clk   => clk,
		reset => cpu_rst,
		AB    => cpu_adr,
		DI    => cpu_din,
		DO    => cpu_dout,
		WE    => cpu_wr,
		IRQ   => cpu_irq,
		NMI   => cpu_nmi,
		RDY   => cpu_rdy,
		SYNC	=> cpu_sync,
		SO		=>	cpu_so,
		isRTS	=> is_rts,
		isJSR	=> is_jsr
	);

	-- on-chip memory, 8 bits wide
	inst_ram64kb: entity work.ram_dual_64k 
	port map (

		-- access from internal 65C02
	  clka    => clk,
	  wea(0)	 => cpu_wr,	
	  ena		 => ram_en,				-- enabled for internal memory access only
	  addra   => cpu_adr,
	  dina    => cpu_dout,
	  douta   => mem_dout,

	  -- access from initialization state machine
	  clkb    => clk,
	  web(0)	 => wr_init,
	  enb		 => en_init,			-- enabled all the time (shadow external memory access)
	  addrb   => ext_a_buf,
	  dinb    => ext_d_i,
	  doutb   => open
	);
	
	-- address decoder for regular operation
	inst_decoder: entity work.decoder
	port map (
		clk		=> clk,
		switch	=> map_i,					-- select memory model
		adr		=> cpu_adr,					-- current memory address
		data		=> cpu_dout,				-- data from CPU (to track bank switching for some hosts)
		write		=> cpu_wr,					-- some memory is external only when writing (screen buffer)

		is_ext	=> is_ext,					-- this address is external (not fast RAM/ROM)	
		is_slow 	=>	is_slow,					-- this address triggers slow (time critical) mode  
		idle		=> idle_page,				-- put this address on the external bus while busy internally
		
		init_adr	=> open,
		init_data => open

	);

	-- address decoder for initialization loop
	inst_decoder_init: entity work.decoder
	port map (
		clk		=> '0',						-- disable slow mode detection
		switch	=> map_i,					-- select memory model
		adr		=> a_init,					-- current memory address - initialization counter
		data		=> x"00",					-- no data input needed
		write		=> '0',						-- we want to read cached memory areas into BRAM, don't consider them external here
		
		is_ext	=> is_ext_init,			-- this address is external (not fast RAM/ROM)	
		is_slow 	=>	open,						-- this address triggers slow (time critical) mode 
		idle		=> open,
		
		init_adr	=> a_init_write,			-- address for host initialization prior to reading RAM/ROM
		init_data => d_init_write			-- data for host initialization
	);
	
	-- data multiplexer
	inst_mux: entity work.datamux
	port map (
		sel_ext_i 	=> was_ext or slow_mode,
		int_d_i		=> mem_dout,
		ext_d_i		=> ext_d_buf,			-- latched data from external bus 
		d_o			=> cpu_din
	);

	-- bidirectional external databus
	inst_databus: entity work.databus 
	port map ( 
		-- external bi-directional bus
		d_io			=> ext_d_io,		
		-- internal bus and direction bit
		d_read_o		=> ext_d_i,				-- data coming in from bidir bus
		d_write_i	=> ext_d_o,				-- data to bidir bus
		wr_n_i		=> not ext_drive_buf	-- '0' to write to bidir bus - latched 
	);
	
	-- override DIP settings via JTAG TDI
   BSCAN_SPARTAN6_inst : BSCAN_SPARTAN6
   generic map (
      JTAG_CHAIN => 1  						-- Value for USER command. Possible values: (1,2,3 or 4).
   )
   port map (
      CAPTURE 	=> open, 					-- 1-bit output: CAPTURE output from TAP controller.
      DRCK 		=> open,       			-- 1-bit output: Data register output for USER functions.
      RESET 	=> open,     				-- 1-bit output: Reset output for TAP controller.
      RUNTEST 	=> open, 					-- 1-bit output: Output signal that gets asserted when TAP controller is in Run Test Idle state.

      SEL 		=> open,         			-- 1-bit output: USER active output.
      SHIFT 	=> open,     				-- 1-bit output: SHIFT output from TAP controller.
      TCK 		=> open,        			-- 1-bit output: Scan Clock output. Fabric connection to TAP Clock pin.
      TDI 		=> map_enable,         	-- 1-bit output: TDI output from TAP controller.
      TMS 		=> open,         			-- 1-bit output: Test Mode Select output. Fabric connection to TAP.
      UPDATE 	=> open,   					-- 1-bit output: UPDATE output from TAP controller
      TDO 		=> '1'          			-- 1-bit input: Data input for USER function.
   );
	
---------------------------------------------------------------------
-- interface between external and internal bus
---------------------------------------------------------------------

-- Registered version of is_ext address decoder.
-- For data input multiplexer, active in the cycle *after* the address is presented.
-- Freeze the was_ext flag while the CPU is halted to wait for an external bus cycle to complete.

process (clk)
begin
   if rising_edge(clk) then
		if rdy_int = '1' then
			was_ext <= is_ext;
		end if;
	end if;        
end process;

-- synchronize asynchronous input signals

process (clk)
begin
	if falling_edge (clk) then
		meta_phi0	<= ext_phi0_i;
		meta_irq_n 	<= ext_irq_n_i;
		meta_nmi_n 	<= ext_nmi_n_i;
		meta_rdy 	<= ext_rdy_i;
		meta_rst_n 	<= ext_rst_n_i;
		meta_so_n 	<= ext_so_n_i;
	end if;
   if rising_edge(clk) then
		sync_phi0	<= meta_phi0;
		sync_irq_n	<= meta_irq_n;
		sync_nmi_n	<= meta_nmi_n;	
		sync_rdy		<= meta_rdy;	
		sync_rst_n	<= meta_rst_n;
		sync_so_n	<= meta_so_n;
	end if;        
end process;


-- Clock outputs
--
-- These will be up to 2.5 clk cycles behind phi0. 	### measured 3.5 cycles -- why?
-- (0..1 clk phi0 to meta_phi0, 0.5 clk meta_phi0 to sync_phi0, 1 clk sync_phi0 to phi2).
-- Spec limit for 4 MHz 65C02 is 30 ns, so we violate the spec for clk < 83 MHz.
-- In a clean host design, this should not matter at all since phi0 should not be used by the host.
--
-- Delay from sync_phi0 to ext_phi2 is needed in order to generate the "look-ahead" edge flags...
-- Note that the "hold data upon read" requirement, tHR, is only 10 ns for 65C02,
-- so we need to read within 10ns after falling edge of Phi2.

process (clk)
begin
   if rising_edge(clk) then
		ext_phi1_o		<= (not sync_phi0);	
		ext_phi2_o		<= sync_phi0;
	end if;        
end process;


-- Delayed versions of edge-sensitive inputs for edge detection,
-- and delayed phi2 for hold timing

process (clk)
begin
   if rising_edge(clk) then
		prev0_phi2	<= sync_phi0;
		prev1_phi2	<= prev0_phi2;
		prev2_phi2	<= prev1_phi2;
		prev3_phi2	<= prev2_phi2;
		prev4_phi2	<= prev3_phi2;
		prev5_phi2	<= prev4_phi2;
		prev6_phi2	<= prev5_phi2;
		prev7_phi2	<= prev6_phi2;
	end if;        
end process;

--	Edge detection signals.
-- No further delay vs. the phi signals please, hence these are asynchronous signals.
-- NOTE: All edge signals must occur at different clk cycles to make the state machine work!

		edgep_phi2	<= not (prev0_phi2) and (sync_phi0);				-- positive edge on phi2: enable data output driver
		edgen_phi2	<= (prev0_phi2) and not (sync_phi0);				-- negative edge on phi2
		edgen_phi2_ahold	<= (prev4_phi2) and not (prev3_phi2);		-- RW and addresses change here: phi2- plus 4*10ns clk (matches R65C02G)
		edgen_phi2_dhold	<= (prev5_phi2) and not (prev4_phi2);		-- data held until here: phi2- plus 5*10ns clk (matches R65C02G)
																						-- this MUST occur after ahold and before phi2 rising edge! 

-- External bus interface.
--
-- Timing requirements from the R65C02 datasheet:
--
-- Address, R/W and SYNC signals are presented shortly after Phi2 falling edge.
-- Data (write) are only presented during Phi2 = H. Keep high-impedance output during Phi2=L to avoid collisions.
-- Date (read) are read on falling edge of Phi2.
--
-- At the end of the external bus cycle (phi2 falling edge), hold times apply:
-- hold write data 30ns, hold rw and sync 15ns.
--
-- We can start a new bus cycle at the same time; that easily gives us enough address setup time (t_acc).
-- this allows for directly consecutive external bus cycles if required.

process (clk)
begin
   if rising_edge(clk) then

		-- FSM to initialize RAM, drive external bus cycles

		case bus_state is
		
			-- power-on reset and initialization
			
			when bus_init_waitreset =>						-- wait for power-on reset to be released and new bus cycle to start
				if sync_rst_n = '1' and edgen_phi2_ahold = '1' then
					rst_n_int <= '0';							-- keep (or put) the internal cpu in reset 
					ext_sync_o <= '0';						-- no meaningful SYNC output
					ext_wr_n_o <= '0';						-- start one write cycle
					ext_drive_buf <= '1';					-- enable data bus output driver		
					ext_a_buf <= a_init_write;
					ext_d_o <= d_init_write;
					bus_state <= bus_init_waitwrite;		-- wait for phi2 low edge
				end if;

			when bus_init_waitwrite =>
				if edgen_phi2_ahold = '1' then			-- wait for completion of prior bus cycle
					ext_wr_n_o <= '1';						-- only reading during further initialization
					ext_drive_buf <= '0';					-- disable Data bus output driver		
					a_init <= x"0000";						-- initialize address for copying external RAM
					bus_state <= bus_init_wait;			-- wait for phi2 low edge
				end if;
							
			when bus_init_wait =>							-- wait for beginning of new active bus cycle
				if edgen_phi2_ahold = '1' then			-- xx ns after phi2 falling edge: new cycle begins
					if is_ext_init = '0' then				-- only non-peripheral addresses go onto the bus (are copied into memory)
						ext_a_buf <= a_init;					-- set address output
					end if;
					a_init <= a_init+1;						-- counter moves on
					bus_state <= bus_init_active;
				end if;
			
			when bus_init_active =>							-- wait for end of active bus cycle
																	-- no need to set ext_drive_o here, as init only *reads* from the external bus
				if edgen_phi2 = '1' then					-- falling phi2 edge.
																	-- wr_init will be triggered at edgen_phi2, see async signals below
					if sync_rdy = '1' then					-- act only if the host is not delaying us
						if a_init = x"0000" then 			-- cycled through complete memory
							rst_n_int <= '1';					-- restart the cpu. 
							bus_state <= bus_idle;			-- drop into normal operation
						else
							bus_state <= bus_init_wait;	-- next address to read
						end if;
					end if;
				end if;
			
			-- regular operation
			
			when bus_idle =>									-- currently no external access
			
				if edgen_phi2_ahold = '1' then			-- terminate prior bus cycle, if any: end of adr & r/wn hold time
					ext_wr_n_o <= '1';						-- reset r/wn output (read)
					ext_a_buf <= idle_page & x"00";		-- set "harmless" address, to avoid multiple triggers to peripheral addresses
					ext_sync_o <= '0';						-- reset sync output (no opcopde fetch)
				end if;
				
				if edgen_phi2_dhold = '1' then			-- end of data hold time
					ext_drive_buf <= '0';					-- disable Data bus output driver		
				end if;

				if (is_ext = '1') or (slow_mode = '1') then							
									
					-- New external bus cycle begins.
					-- slow_mode (executing time-critical code, but accessing internal RAM) is handled just like is_ext.
					--
					-- Writing always goes to external bus *and* internal RAM.
					-- Reading must come from internal RAM when slow_mode AND NOT is_ext;
					-- this is achieved by latching internal RAM data in ext_d_buf in the *following* cycle.
					-- CPU's input data multiplexer must select ext_d_buf when slow_mode OR was_ext.
					
					rdy_int <= '0';							-- stop cpu. cpu outputs may change before it stops, so register them below...
					cpu_wr_latched <= cpu_wr;				-- latch current CPU status, but do not output on external bus yet
					cpu_adr_latched <= cpu_adr;			-- ... because the previous cycle's hold time might still be ongoing
					cpu_dout_latched <= cpu_dout;			-- ...
					cpu_sync_latched <= cpu_sync;			-- ... 
					new_cycle <= '1';							-- next clk tick: latch is_slow and mem_dout
					bus_state <= bus_wait_ahold;			-- wait for end of adr hold time
				end if;

			when bus_wait_ahold =>							-- wait for end of previous adr hold time

				-- during 1st clk tick in this state, deal with slow_mode
				if new_cycle = '1' then						-- first clk tick in this state
					new_cycle <= '0';
					ext_d_buf <= mem_dout;					-- data from BRAM; present to the CPU at end of slow_mode bus cycle

					if is_slow = '1' then					-- critical I/O address triggers slow mode
																	-- (is_slow output in decoder is registered, 
																	-- hence refers to prior clk's address!)
						if slow_mode = '0' then				-- enter slow_mode
							slow_mode <= '1';							
							rts_level <= 0;
							slow_timeout <= 1; 
						elsif rts_level = 0 then			-- slow_mode ongoing, access at top stack level: 
							slow_timeout <= 1; 				-- re-start timeout counter
						end if;
					else
						-- count cycles without access to slow I/O,
						-- end slow mode after timeout (for cases where slow I/O is access in top level code)
						slow_timeout <= slow_timeout+1;	
						if slow_timeout = 0 then			-- timer has overflowed, timeout
							slow_mode <= '0';					-- reset slow mode
						end if;
					end if;

					-- monitor instructions which control the end of slow mode
					if slow_mode = '1' then					-- operating in slow mode: track RTI/RTS/JSR
						if is_rts = '1' then					-- RTS or RTI
							if rts_level = 0 then 
								slow_mode <= '0';				-- returning from level where slow mode was first started 
							else 
								rts_level <= rts_level - 1;-- track subroutine level
							end if;
						end if;
						if is_jsr = '1' then 
							rts_level <= rts_level + 1;	-- JSR: track subroutine levels and BRK/IRQ/NMI
						end if;
					end if;
				end if;

				-- at end of previous address hold time, set external bus signals
				if edgen_phi2_ahold = '1' then			-- address hold time reached
					ext_a_buf <= cpu_adr_latched;			-- time to set the outputs
					ext_wr_n_o <= not cpu_wr_latched;	-- read or write access?
					ext_sync_o <= cpu_sync_latched;					
					bus_state <= bus_wait_dhold;
				end if;
				
			when bus_wait_dhold =>							-- wait for end of data hold time
				if edgen_phi2_dhold = '1' then			-- data hold time reached
					ext_d_o <= cpu_dout_latched;			-- write data, but output driver remains disabled for now
					ext_drive_buf <= '0';					-- disable Data bus output driver		
					bus_state <= bus_wait_write;
				end if;

			when bus_wait_write =>							-- wait for rising phi2 edge
				if edgep_phi2 = '1' then					-- *rising* phi2 edge 
					ext_drive_buf <= cpu_wr_latched;		-- enable Data bus output driver during write cycles
					bus_state <= bus_wait_read;
				end if;

			when bus_wait_read =>							-- wait for falling phi2 edge
				if edgen_phi2 = '1' then					-- *falling* phi2 edge, end of this external cycle
					if sync_rdy = '1' then					-- act only if the host is not delaying us
						if was_ext = '1' then				-- in slow_mode, RAM data are already in ext_d_buf!
							ext_d_buf <= ext_d_i;			-- get data now -- don't require hold time from host
						end if;
						rdy_int <= '1';						-- restart the cpu. ext_d_o will remain valid for dhold time 
						bus_state <= bus_idle;				-- hold time extends into next cpu cycle(s)
					end if;
				end if;
				
		end case;
		
	end if;
end process;

	-------------------------
	-- Status output & test signals.
	-- Latch user interface signals to relax timing/routing
	-------------------------

process (clk)
begin
   if rising_edge(clk) then

		-- LED brightness reflects level of acceleration.
		-- Blink LED off during RAM initialization to indicate that configuration has worked.
		led_latch <= rst_n_int and (not was_ext) and (not slow_mode);
		led <= led_latch;
		
		-- memory map selector. TDI=0 overrides DIP switches.
		if map_enable = '1' 
		then map_i <= not map_n_i;
		else map_i <= "0000";
		end if;
		
		-- Test outputs, go to unused DIP pins with direct 3.3V output level. 
		-- test(0) = DIP pin 36, test (1) = DIP pin 35.
		-- Set test (0) = '1' or 'Z' when not in use (has pullup to 3.3V).
		test_latch (0) <= is_slow;
		test_latch (1) <= slow_mode;
--		test_latch (0)	<= is_rts;
--		if bus_state = bus_wait_ahold then test_latch (1) <= edgen_phi2_ahold;
--		else test_latch (1) <= '0';
--		end if;
		test <= test_latch;
		
	end if;
end process;

	-------------------------
	-- external bus, async signals
	-------------------------

	-- control RAM port B (active during initialization
	-- wr_init will write data to block RAM in the peripheral address space too,
	-- since we cycle through each a_init address. Doesn't matter.
	
	en_init <= '1' when (bus_state=bus_init_active) else '0';
	wr_init <= edgen_phi2 when (sync_rdy='1') and bus_state=bus_init_active else '0';

	-- control signal inputs to softcore CPU.
	
	cpu_rst	<= not (sync_rst_n and rst_n_int); 	-- external reset or internal signal from init loop can pull low
	cpu_irq	<= not sync_irq_n;
	cpu_nmi	<= not sync_nmi_n;
	cpu_so	<= not sync_so_n;
	cpu_rdy	<= rdy_int;								-- external RDY acts via bus FSM; only internal signal acts directly on CPU

--	ram_en	<= not cpu_rst;						-- enable internal RAM ### fast, but reliability issues with Milano??
	ram_en	<= not cpu_rst and rdy_int;		-- enable internal RAM ### works well in Milano, should work with slow_mode 
	
	-- external output signals: 
	-- echo internal shadow copies
	ext_dwrite_n_o	<= not ext_drive_buf;		-- Rev C/D: enable output driver 
	ext_dread_n_o	<= ext_drive_buf;				-- Rev C/D: enable input buffer
	ext_a_o 			<= ext_a_buf;					-- address bus
		
end BEHAVIORAL;